import asyncio
import datetime
import importlib
import logging
from typing import Callable

import celery
import pytz
import uvicorn
from celery import Celery
from celery.backends.database import session_cleanup
from rich import box
from rich.live import Live
from rich.logging import RichHandler
from rich.panel import Panel
from rich.progress import Progress, TextColumn, BarColumn
from rich.table import Table
from scheduler.asyncio import Scheduler
from scheduler.asyncio.job import Job
from scheduler.trigger.core import Weekday
from sqlalchemy import select
from sqlalchemy.orm import joinedload

from orchestra.formatting import pretty_print_block, get_job_state
from orchestra.models import Run, TimingAwareTask
from orchestra.scheduling import Schedule

logging.basicConfig(
    level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler(rich_tracebacks=True, tracebacks_show_locals=True)]
)

logger = logging.getLogger("orchestra.core")

instance: "Orchestra | None" = None


class Orchestra(Celery):
    def __init__(self, backend_conn_str: str, backend_module_name: str = "orchestra.backend.OrchestraBackend", enable_api=False, api_address: str = "localhost:5000",
                 scheduler_loop_resolution_in_seconds: float | int = 0.1, **kwargs):
        super().__init__(
            backend=f"{backend_module_name}+{backend_conn_str}",
            result_extended=True,
            task_track_started=True,
            task_send_sent_event=True,
            **kwargs,
        )
        self.scheduler: Scheduler | None = None
        self.loop = None
        self.enable_api = enable_api
        self.api_address = api_address
        self.loop_resolution_in_seconds = scheduler_loop_resolution_in_seconds
        self.server: uvicorn.Server | None = None
        self.paused_jobs: set[Job] = set()

    def get_event_loop(self, loop):
        self.loop = loop or asyncio.get_running_loop()

    async def api_server(self):
        log_config = uvicorn.config.LOGGING_CONFIG
        log_config["loggers"] = []

        try:
            host, port_str = self.api_address.split(":")
            port = int(port_str)
        except ValueError:
            host = self.api_address
            port = 8000

        config = uvicorn.Config("orchestra.api.main:app", host=host, port=port, log_config=log_config)

        self.server = uvicorn.Server(config)
        await self.server.serve()

    async def live_output(self):
        with Live(
                self.get_scheduler_status_table(), refresh_per_second=10
        ) as live:
            while True:
                await asyncio.sleep(self.loop_resolution_in_seconds)
                live.update(self.get_scheduler_status_table())

    async def run(self) -> None:
        logger.info("Orchestra starting")

        # we need to run 2 tasks in parallel
        # first one is the live output on the CLI/TUI:sdf
        tasks = [asyncio.create_task(self.live_output())]

        # the second one is the optionally enabled API server
        if self.enable_api:
            tasks.append(asyncio.create_task(self.api_server()))
            global instance
            instance = self
        try:
            # uvicorn swallows Cancellation requests >>>:((((
            # https://github.com/encode/uvicorn/issues/1579
            # https://docs.python.org/3/library/asyncio-task.html#task-cancellation
            # we have to exit when either uvicorn (api server) or the Orchestra scheduler stops
            _, pending_tasks = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)

            # if we get to this point then one of the above tasks were cancelled by the user
            logger.warning("User-requested shutdown.")
            for task in pending_tasks:
                task.cancel()

            await asyncio.wait(pending_tasks)

        except (asyncio.CancelledError, KeyboardInterrupt):
            logger.warning("User-requested shutdown.")

    def get_job_by_name(self, job_name: str, include_paused: bool = False) -> Job | None:
        for job in self.scheduler.jobs.union(self.paused_jobs if include_paused else []):
            if job.alias == job_name:
                return job

    def wrap_celery_task(self, job_name: str, task: celery.Task):
        async def celery_task():
            task_meta = task.delay()
            logger.debug(f"Triggered job {job_name}")
            task_meta.created_date = datetime.datetime.now(tz=pytz.utc)

            session = self.backend.ResultSession()
            with session_cleanup(session):
                trigger_timestamp = datetime.datetime.now(tz=pytz.utc)
                job = self.get_job_by_name(job_name, include_paused=True)
                if job is None:
                    logger.critical(f"Unknown job with name {job_name}")

                module_name, _, task_name = task.name.rpartition(".")

                scheduler_log = Run(
                    job=job_name,
                    module=module_name,
                    task=task_name,
                    task_id=task_meta.id,
                    schedule=job.type.name if job.max_attempts != 1 else "ONCE",
                    timezone=job.datetime.tzname(),
                    triggered_date=trigger_timestamp,
                )
                session.add(scheduler_log)
                session.commit()

        celery_task.__name__ = task.name
        return celery_task

    def __create_job_from_celery_task_schedule(
            self, job_name: str, module_name: str, task_name: str) -> Callable:
        module = importlib.import_module(module_name)
        task: celery.Task = getattr(module, task_name)

        return self.wrap_celery_task(job_name, task)

    def schedule_celery_task(self, job_name: str, task: Callable, schedule: Callable[..., Job], timing: datetime.datetime | datetime.timedelta | datetime.time | Weekday,
                             tags: set[str] | None = None, attempt_resume: bool = False):
        module_name, _, task_name = task.name.rpartition(".")  # type:ignore
        self.schedule_job(job_name=job_name, module_name=module_name, task_name=task_name, schedule=schedule, timing=timing, tags=tags, attempt_resume=attempt_resume)

    async def create_schedule(self, module_definitions: list[dict] = None, loop=None) -> None:
        self.scheduler = self.scheduler or Scheduler(tzinfo=pytz.utc, loop=self.get_event_loop(loop))
        if module_definitions is not None:
            self.add_schedules(module_definitions)

    def schedule_job(self, job_name: str, module_name: str, task_name: str, schedule: Callable[..., Job], timing: datetime.datetime | datetime.timedelta | datetime.time | Weekday,
                     tags: set[str] | None = None, attempt_resume: bool = False):
        session = self.backend.ResultSession()
        with session_cleanup(session):
            resume_parameters: dict = {}
            if attempt_resume:
                logs = (
                    select(Run)
                    .where(Run.job == job_name)
                    .order_by(Run.triggered_date.desc())
                    .limit(1)
                )
                last_run: Run | None = next(session.scalars(logs), None)
                if last_run is not None:
                    last_run_utc = last_run.triggered_date.replace(tzinfo=pytz.utc)
                    last_running_time_local = last_run_utc.astimezone(
                        pytz.timezone(last_run.timezone)
                    )
                    match last_run.schedule:
                        case "ONCE":
                            logger.warning(
                                f"Job {job_name} was scheduled to run exactly once, but it has already ran on {last_running_time_local}, tz={last_run.timezone}, not scheduling again."
                            )
                            return None

                        case "CYCLIC":
                            utc_now = datetime.datetime.now(tz=pytz.utc)
                            next_run_utc = last_run_utc + datetime.timedelta(
                                seconds=((utc_now - last_run_utc).total_seconds() // timing.total_seconds()) * timing.total_seconds() + timing.total_seconds()
                            )
                            next_run_local = next_run_utc.astimezone(
                                pytz.timezone(last_run.timezone)
                            )

                            resume_parameters = {
                                "start": next_run_local - datetime.timedelta(seconds=timing.total_seconds())
                            }

            schedule(
                timing=timing,
                handle=self.__create_job_from_celery_task_schedule(
                    job_name, module_name, task_name
                ),
                alias=job_name,
                tags=tags or set(),
                **resume_parameters,
            )
            if "start" in resume_parameters:
                logger.warning(
                    f"Job {job_name} was scheduled to run {timing}, resuming using {last_running_time_local}, tz={last_run.timezone} as reference time."
                    f" Next run at {next_run_local}, tz={last_run.timezone}"
                )
            else:
                logger.info(f"Job {job_name} was scheduled to run {timing}")

    def add_schedules(self, module_definitions: list[dict], attempt_resume: bool = True):
        for index, block in enumerate(module_definitions or []):
            block_name: str = block.get("name", f"{index + 1}. schedule block")
            block_module: str | None = block.get("module")

            if block_module is None:
                error_message: str = f"Module is undefined for {block_name}. Definition:\n\n{pretty_print_block(block)}"
                logger.error(error_message)
                raise ValueError(error_message)

            schedule_definitions: dict | None = block.get("schedules")

            if schedule_definitions is None:
                error_message: str = f'Missing schedule definition for "{block_name}" module. Module definition:\n\n{pretty_print_block(block)}'
                logger.error(error_message)
                raise ValueError(error_message)

            for task_index, schedule_definition in enumerate(schedule_definitions):
                if "name" not in schedule_definition:
                    logger.critical(f"Missing job name in schedule in {block_name} module. Job definition:\n\n{pretty_print_block(schedule_definition)}")
                    continue

                job_name: str = schedule_definition.get("name")

                if not schedule_definition.get("enabled", False):
                    logger.info(f"Ignoring disabled job {job_name} in {block_name}. Job definition:\n\n{pretty_print_block(schedule_definition)}")
                    continue

                task_name: str = schedule_definition.get("task")
                if task_name is None:
                    logger.error(f"Task is undefined for {job_name} in {block_name}. Definition:\n\n{pretty_print_block(schedule_definition)}")
                    continue

                if self.get_job_by_name(job_name) is not None:
                    logger.critical(f"Can't add job with name {job_name}, because a job with that name already exists")
                    continue

                if not schedule_definition.get("schedule", False):
                    logger.critical(f'Missing schedule definition for enabled Job "{block_name}". Job definition:\n\n{pretty_print_block(schedule_definition)}')
                    continue

                schedule = Schedule(
                    definition=schedule_definition, scheduler=self.scheduler
                )

                self.schedule_job(job_name=job_name,
                                  module_name=block_module,
                                  task_name=task_name,
                                  schedule=schedule.job_type,
                                  timing=schedule.get_timing(),
                                  tags=set(schedule_definition.get("tags", set())),
                                  attempt_resume=attempt_resume)

    @classmethod
    def set_job_state_property(cls, job: Job, is_paused: bool) -> Job:
        job.is_paused = is_paused
        return job

    def apply_state_to_jobs(self, jobs: set[Job]) -> set[Job]:
        return set([self.set_job_state_property(job, is_paused=True if job in self.paused_jobs else False) for job in jobs])

    def get_jobs(self, tags: set[str], any_tag: bool, include_paused: bool = True):
        # contrary to what get_jobs' docstring says, get_job will not return all jobs when tag is an empty set
        if len(tags) == 0:
            active_jobs = self.scheduler.get_jobs()
        else:
            active_jobs = self.scheduler.get_jobs(tags, any_tag=any_tag)
        paused_jobs = set()
        if include_paused:
            paused_jobs = self.get_paused_jobs(tags, any_tag=any_tag)

        return self.apply_state_to_jobs(active_jobs.union(paused_jobs))

    def job_exists(self, job_name: str):
        return self.get_job_by_name(job_name, include_paused=True) is not None

    def get_paused_jobs(self, tags: set[str], any_tag: bool):
        if len(tags) == 0:
            return self.paused_jobs
        return self.apply_state_to_jobs(set(filter(lambda job: job.tags.intersection(tags) if any_tag else job.tags == tags, self.paused_jobs)))

    def pause_job(self, job_name: str):
        job = self.get_job_by_name(job_name)
        assert job is not None
        self.scheduler.delete_job(job)
        self.paused_jobs.add(job)
        logger.info(f"Paused job {job_name}")
        return self.set_job_state_property(job, is_paused=True)

    def pause_jobs_with_tags(self, tags: set[str], any_tag: bool):
        jobs_to_pause = self.get_jobs(tags, any_tag, include_paused=False)

        for job in jobs_to_pause:
            self.pause_job(job.alias)

        return self.apply_state_to_jobs(jobs_to_pause)

    def resume_job(self, job_name: str) -> Job:
        job = self.get_job_by_name(job_name, include_paused=True)
        assert job is not None
        self.paused_jobs.remove(job)
        task = self.scheduler._Scheduler__loop.create_task(self.scheduler._Scheduler__supervise_job(job))
        self.scheduler._Scheduler__jobs[job] = task
        logger.info(f"Resumed job {job_name}")
        return self.set_job_state_property(job, is_paused=False)

    def resume_jobs_with_tags(self, tags: set[str], any_tag: bool):
        jobs_to_resume = self.get_paused_jobs(tags, any_tag)

        for job in jobs_to_resume:
            self.resume_job(job.alias)

        return self.apply_state_to_jobs(jobs_to_resume)

    def delete_job(self, job_name: str) -> Job | None:
        job = self.get_job_by_name(job_name, True)
        if job in self.paused_jobs:
            self.paused_jobs.remove(job)
        else:
            self.scheduler.delete_job(job)

        return job

    def delete_jobs_with_tags(self, tags: set[str], any_tag: bool, include_paused: bool = True):
        active_jobs_to_delete = self.apply_state_to_jobs(self.get_jobs(tags, any_tag, include_paused=False))
        self.scheduler.delete_jobs(tags, any_tag)

        paused_jobs_to_delete = set()
        if include_paused:
            paused_jobs_to_delete = self.apply_state_to_jobs(self.get_jobs(tags, any_tag, include_paused=True))
            for job in paused_jobs_to_delete:
                self.delete_job(job.alias)
        return active_jobs_to_delete.union(paused_jobs_to_delete)

    async def trigger_job(self, job_name: str):
        job = self.get_job_by_name(job_name, include_paused=True)
        assert job is not None
        await job.handle()
        logger.info(f"Job {job_name} triggered manually")
        if job.max_attempts == 1:
            self.scheduler.delete_job(job)

    async def trigger_jobs_with_tags(self, tags: set[str], any_tag: bool, include_paused: bool):
        jobs_to_trigger = self.apply_state_to_jobs(self.get_jobs(tags, any_tag, include_paused))

        async with asyncio.TaskGroup() as tg:
            [tg.create_task(job.handle()) for job in jobs_to_trigger]

        for job in jobs_to_trigger:
            logger.info(f"Job {job.alias} triggered manually")
            if job.max_attempts == 1:
                self.scheduler.delete_job(job)

        return jobs_to_trigger

    def get_runs_of_a_job(self, job_name: str, page_size: int, page: int) -> list[Run]:
        session = self.backend.ResultSession()
        with session_cleanup(session):
            runs: list[Run] = list(session.scalars(select(Run).where(Run.job == job_name).order_by(Run.triggered_date.desc()).offset((page - 1) * page_size).limit(page_size).options(joinedload(Run.task_object))))
            return runs

    def get_run_of_a_job_by_id(self, job_name: str, run_id: int) -> Run | None:
        session = self.backend.ResultSession()
        with session_cleanup(session):
            run: Run = next(session.scalars(select(Run).where(Run.job == job_name, Run.id == run_id).limit(1).options(joinedload(Run.task_object))), None)
            return run

    def get_task_by_id(self, task_id: str) -> TimingAwareTask | None:
        session = self.backend.ResultSession()
        with session_cleanup(session):
            return next(session.scalars(select(TimingAwareTask).where(TimingAwareTask.task_id == task_id).limit(1)), None)

    def get_scheduler_status_table(self) -> Table:
        grid = Table.grid(expand=True)
        grid.add_column()

        table = Table(expand=True)
        table.border_style = "bright_yellow"
        table.box = box.ROUNDED
        table.pad_edge = False

        table.add_column("State")
        table.add_column("Shedule", style="green")
        table.add_column("Job name", style="blue")
        table.add_column("Module and task", style="magenta")
        table.add_column("Due at", style="red")
        table.add_column("Timezone", style="bright_blue")
        table.add_column("Due in", style="cyan")
        table.add_column("Attempts")
        table.add_column("Tags")

        for job in self.scheduler.jobs:
            table.add_row(*(["[green]Running[/]"] + get_job_state(job)))

        for job in self.paused_jobs:
            table.add_row(*(["[yellow]Paused[/]"] + get_job_state(job)))

        progress = Progress(TextColumn("{task.description}"), BarColumn(bar_width=None), expand=True, transient=True)
        progress.add_task("Orchestrating jobs", total=None)
        grid.add_row(table)
        grid.add_row(Panel(progress, expand=True, border_style="bright_yellow"))

        return grid



import logging
import re
import sys
from datetime import datetime, time, timedelta
from typing import Callable, Optional

import pytz.exceptions
import scheduler.trigger as weekdays
from pytz import timezone
from scheduler import Scheduler

from orchestra.formatting import pretty_print_block, transform_timing_to_schedule

logger = logging.getLogger("orchestra.scheduling")


class Schedule:
    job_type: Callable
    trigger: Optional[Callable]
    timing: time | datetime | timedelta
    timezone: pytz.tzinfo.BaseTzInfo

    def __init__(self, definition: dict, scheduler: Scheduler):
        self.definition = definition
        self.scheduler = scheduler
        if "schedule" not in definition:
            logger.critical(
                f"Missing schedule definition in:\n\n{pretty_print_block(self.definition)}"
            )
        if "timing" not in definition["schedule"]:
            logger.critical(
                f"Missing timing definition in:\n\n{pretty_print_block(self.definition)}"
            )

        self.job_type, self.trigger, self.timing = self.parse_timing()
        self.timezone = self.maybe_parse_timezone()

    @staticmethod
    def clean_timing(timing_string: str) -> str:
        """
        - Remove all "on"-s with "at"-s
        - Remove all commas
        - Replace all "next" with "once at"
        - Remove any "and"-s
        - Replace all "X hour Y minutes Z seconds" with "XX:YY:ZZ"
        - Replace all "day at XX" with "00:00:XX"
        - Replace all "XX:YY" with "00:XX:YY
        - Remove all double spaces
        """

        timing_string = timing_string.replace(" on ", " at ")
        timing_string = timing_string.replace(",", "")
        timing_string = timing_string.replace("next", "once at")
        timing_string = timing_string.replace("and", "")

        seconds = "00"
        minutes = "00"
        hours = "00"
        normalize_time = False

        # remove any mention of seconds and remember the value
        if (match := re.search(r"(\d+)\s?seconds?", timing_string)) is not None:
            seconds = match.group(1)
            timing_string = timing_string.replace(match.group(0), "")
            normalize_time = True

        # remove any mention of minutes and remember the value
        if (match := re.search(r"(\d+)\s?minutes?", timing_string)) is not None:
            minutes = match.group(1)
            timing_string = timing_string.replace(match.group(0), "")
            normalize_time = True

        # remove any mention of hours and remember the value
        if (match := re.search(r"(\d+)\s?hours?", timing_string)) is not None:
            hours = match.group(1)
            timing_string = timing_string.replace(match.group(0), "")
            normalize_time = True

        # convert every hour|day [at] X:Y to pre or post zero-padded XX:YY:ZZ format, based on if timing is hourly or daily respectively
        if (
            match := re.search(
                r"\b(day|hour)(?:\Wat\W)?(\d+):(\d{2})(?!:)", timing_string
            )
        ) is not None:
            normalize_time = True
            timing_string = timing_string.replace(match.group(0), match.group(1))
            match match.group(1):
                case "day":
                    hours = match.group(2)
                    minutes = match.group(3)
                    seconds = "00"
                case "hour":
                    hours = "00"
                    minutes = match.group(2)
                    seconds = match.group(3)

        # normalize bare timestamps like XX:YY to XX:YY:ZZ format
        if (
            match := re.search(r"^(\d{1,2}):(\d{1,2})(?::(\d{1,2}))?$", timing_string)
        ) is not None:
            normalize_time = True
            hours = match.group(1)
            minutes = match.group(2)
            seconds = match.group(3) if match.group(3) else "00"
            timing_string = timing_string.replace(match.group(0), "once at")

        # normalize bare timestamps like YYYY-mm-dd XX:YY to YYYY-mm-dd XX:YY:ZZ format
        if (
            match := re.search(
                r"^(\d{4}-\d{2}-\d{2}) (\d{1,2}):(\d{1,2})$", timing_string
            )
        ) is not None:
            normalize_time = True
            hours = match.group(2)
            minutes = match.group(3)
            seconds = "00"
            timing_string = timing_string.replace(match.group(0), match.group(1))

        if normalize_time:
            timing_string += f" {int(hours):02}:{int(minutes):02}:{int(seconds):02}"

        while timing_string != timing_string.replace("  ", " "):
            timing_string = timing_string.replace("  ", " ")

        return timing_string

    def parse_timing(self):
        """
        every 00:00:01
        every 48:00:00
        every 31 minutes and 12 seconds
        every Monday at 03:15:00
        next Tuesday at 13:11:00
        once, Wednesday at 01:15:00
        once on Thursday, 00:31:41
        every day at 09:15:14
        every minute at 15 seconds
        every hour at 05:00
        every day on 9:03
        every day at 9:15
        every day at 13 hours
        once in 13:11:52
        11:23:00
        9:13
        2023-04-01 09:13
        2023-04-01 09:13:11
        next 23:00:00
        once in 5 hours 2 minutes and 15 seconds
        once in 1 hour and 10 minutes
        once in 2 minutes and 10 seconds
        once in 17 seconds
        """

        clean_timing = self.clean_timing(self.definition["schedule"]["timing"])

        # there is a complete datetime in the timing, those can only happen once, without a trigger
        if (
            match := re.search(
                r"(\d{4})-(\d{2})-(\d{2})\s(\d{2}):(\d{2}):(\d{2})",
                clean_timing,
            )
        ) is not None:
            return (
                self.scheduler.once,
                None,
                self.get_localized_datetime(
                    year=int(match.group(1)),
                    month=int(match.group(2)),
                    day=int(match.group(3)),
                    hour=int(match.group(4)),
                    minute=int(match.group(5)),
                    second=int(match.group(6)),
                ),
            )

        # this is just a bare time, we schedule the job to run once at the next time specified
        if (
            match := re.search(r"^(?:once at )?(\d{2}):(\d{2}):(\d{2})", clean_timing)
        ) is not None:
            return (
                self.scheduler.once,
                None,
                self.get_localized_time(
                    hour=int(match.group(1)),
                    minute=int(match.group(2)),
                    second=int(match.group(3)),
                ),
            )

        # the matching schedules run once at a specific time, there may be an optional trigger/weekday
        if (
            match := re.search(
                r"^once\s(at|in)?\s?(Monday|Tuesday|Wednesday|Thursday|Friday|Saturday|Sunday)?\s?(?:at)?\s?(\d{2}):(\d{2}):(\d{2})",
                clean_timing,
            )
        ) is not None:
            weekday = match.group(2)
            trigger = getattr(weekdays, weekday) if weekday else None

            # if there's an "in" word after once, then it's a timedelta
            if match.group(1) == "in":
                return (
                    self.scheduler.once,
                    trigger,
                    timedelta(
                        hours=int(match.group(3)),
                        minutes=int(match.group(4)),
                        seconds=int(match.group(5)),
                    ),
                )
            else:
                return (
                    self.scheduler.once,
                    trigger,
                    self.get_localized_time(
                        hour=int(match.group(3)),
                        minute=int(match.group(4)),
                        second=int(match.group(5)),
                    ),
                )

        if (
            match := re.search(
                r"^every\s(?:(day|minute|hour)|(Monday|Tuesday|Wednesday|Thursday|Friday|Saturday|Sunday))?\s?(?:at)?\s?(\d{2}):(\d{2}):(\d{2})",
                clean_timing,
            )
        ) is not None:
            timing = match.group(1)
            weekday = match.group(2)

            # if there's a timing (every day/hour/minute) or a weekday (Monday, Tuesday, etc.) specified the time can only be a specific time
            if timing or weekday:
                trigger = getattr(weekdays, weekday) if weekday else None
                schedule = (
                    getattr(self.scheduler, transform_timing_to_schedule(timing))
                    if timing
                    else self.scheduler.weekly
                )

                return (
                    schedule,
                    trigger,
                    self.get_localized_time(
                        hour=int(match.group(3)),
                        minute=int(match.group(4)),
                        second=int(match.group(5)),
                    ),
                )

            # at this point the timing can only be an interval, e.g.: every 00:00:01, meaning every 1 second, etc.
            return (
                self.scheduler.cyclic,
                None,
                timedelta(
                    hours=int(match.group(3)),
                    minutes=int(match.group(4)),
                    seconds=int(match.group(5)),
                ),
            )

        logger.critical(
            f"Couldn't parse timing definition in:\n\n{pretty_print_block(self.definition)}"
        )

    def maybe_parse_timezone(self) -> pytz.tzinfo.BaseTzInfo | None:
        tz = self.definition.get("schedule").get("timezone")
        if tz is None:
            return pytz.utc
        try:
            return timezone(tz)
        except pytz.exceptions.UnknownTimeZoneError:
            logger.critical(
                f"Unable to parse timezone {tz} in:\n\n{pretty_print_block(self.definition)}"
            )

    def get_timing(self):
        if self.trigger:
            return self.trigger(self.timing)
        else:
            return self.timing

    def get_localized_datetime(
        self, year: int, month: int, day: int, hour: int, minute: int, second: int
    ) -> datetime:
        tz = self.maybe_parse_timezone()
        naive_datetime = datetime(
            year=year, month=month, day=day, hour=hour, minute=minute, second=second
        )
        return tz.localize(naive_datetime)

    def get_localized_time(self, hour: int, minute: int, second: int) -> time:
        tz = self.maybe_parse_timezone()
        return time(hour=hour, minute=minute, second=second, tzinfo=tz)



from typing import Annotated

from fastapi import APIRouter, Path, Response, status

from orchestra.api.dto import TaskDTO
from orchestra.core import instance

router = APIRouter()


@router.get("/tasks/{task_id}", tags=["tasks"])
async def get_task_by_id(task_id: Annotated[str, Path(description="The ID of the task to return")], response: Response) -> TaskDTO | str:
    task = instance.get_task_by_id(task_id)
    if task is None:
        response.status_code = status.HTTP_404_NOT_FOUND
        return f"Task with id {task_id} does not exist"
    else:
        return TaskDTO.map(task)



import datetime

from pydantic import BaseModel
from scheduler.asyncio.job import Job


class JobDTO(BaseModel):
    schedule: str
    job_name: str | None
    task: str
    start: datetime.datetime | None
    due_at: datetime.datetime
    tzinfo: str | None
    due_in: float
    attempts: int
    max_attempts: str | int
    tags: set[str] | None
    is_paused: bool

    @classmethod
    def map(cls, job: Job) -> "JobDTO":
        return JobDTO(schedule=job._BaseJob__type.name.lower(),
                      job_name=job.alias,
                      start=job.start,
                      due_at=job.datetime,
                      tzinfo=str(job.tzinfo),
                      due_in=job.timedelta(datetime.datetime.now(job.tzinfo)).total_seconds(),
                      task=job.handle.__name__,
                      attempts=job.attempts,
                      max_attempts=job.max_attempts,
                      tags=job.tags,
                      is_paused=job.is_paused)  # is_paused is dynamically added to the native Job object



from datetime import datetime
from enum import Enum

import sqlalchemy as sa
from celery.backends.database.models import TaskExtended, ResultModelBase
from sqlalchemy import TIMESTAMP, Text, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship


class TimingAwareTask(TaskExtended):
    """Task result/status."""
    __tablename__ = 'celery_taskmeta'
    __table_args__ = {'sqlite_autoincrement': True, 'extend_existing': True}

    date_created = sa.Column(TIMESTAMP(timezone=True), default=datetime.utcnow)
    date_started = sa.Column(TIMESTAMP(timezone=True), nullable=True)

    def to_dict(self):
        task_dict = super().to_dict()
        task_dict.update({
            'date_created': self.date_created,
            'date_started': self.date_started,
        })
        return task_dict


class Run(ResultModelBase):
    __tablename__ = "orchestra_run"
    __table_args__ = {'sqlite_autoincrement': True}

    id: Mapped[int] = mapped_column(primary_key=True)
    job: Mapped[str]
    module: Mapped[str]
    task: Mapped[str]
    task_id: Mapped[str] = mapped_column(Text, ForeignKey("celery_taskmeta.task_id"))
    schedule: Mapped[str]
    timezone: Mapped[str | None]
    triggered_date: Mapped[datetime] = sa.Column(TIMESTAMP(timezone=True))

    task_object = relationship("TimingAwareTask", foreign_keys=[task_id])


class Status(str, Enum):
    pending = "pending"
    started = "started"
    success = "success"
    failure = "failure"



import logging
from datetime import datetime, timezone

from celery import states
from celery.backends.database import DatabaseBackend

from orchestra.models import TimingAwareTask

logger = logging.getLogger("orchestra.backend")


class OrchestraBackend(DatabaseBackend):

    def __init__(self, dburi=None, engine_options=None, url=None, **kwargs):
        super().__init__(dburi, engine_options, url, **kwargs)

        conf = self.app.conf

        if self.extended_result:
            self.task_cls = TimingAwareTask

        schemas = conf.database_table_schemas or {}
        tablenames = conf.database_table_names or {}
        self.task_cls.configure(
            schema=schemas.get('task'),
            name=tablenames.get('task'))
        self.taskset_cls.configure(
            schema=schemas.get('group'),
            name=tablenames.get('group'))

    def _update_result(self, task, result, state, traceback=None, request=None):

        meta = self._get_result_meta(result=result, state=state, traceback=traceback, request=request, format_date=False, encode=True)

        # Exclude the primary key id, task_id, date_created and date_started columns
        # as we should not set it None
        columns = [column.name for column in self.task_cls.__table__.columns if column.name not in {"id", "task_id", "date_created", "date_started"}]

        status = meta.get("status")
        match status:
            case states.STARTED:
                setattr(task, "date_started", datetime.now(timezone.utc))
                logger.debug(f"Started work on {task.id}")

        # Iterate through the columns name of the table
        # to set the value from meta.
        # If the value is not present in meta, set None
        for column in columns:
            value = meta.get(column)
            setattr(task, column, value)



import json
import datetime
from typing import Any

from pydantic import BaseModel

from orchestra.models import TimingAwareTask, Status


class TaskDTO(BaseModel):
    id: int
    task_id: str
    status: Status
    result: Any
    date_done: datetime.datetime
    traceback: str | None
    args: list
    kwargs: dict
    worker: str
    retries: int
    queue: str
    date_created: datetime.datetime
    date_started: datetime.datetime

    @classmethod
    def map(cls, task: TimingAwareTask) -> "TaskDTO":
        return TaskDTO(id=task.id,
                       task_id=task.task_id,
                       status=Status[task.status.lower()],
                       result=task.result,
                       date_done=task.date_done,
                       traceback=task.traceback,
                       args=json.loads(task.args.decode("utf-8")),
                       kwargs=json.loads(task.kwargs.decode("utf-8")),
                       worker=task.worker,
                       retries=task.retries,
                       queue=task.queue,
                       date_created=task.date_created,
                       date_started=task.date_started)




import yaml
from rich import box
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, BarColumn, TextColumn
from rich.table import Table
from scheduler import Scheduler
from scheduler.asyncio.job import Job


def pretty_print_block(config_block: dict) -> str:
    return yaml.dump(config_block, allow_unicode=True, default_flow_style=False)


def transform_timing_to_schedule(timing: str) -> str:
    """
    timing: "day", "hour", "minute"

    returns daily, hourly, minutely, whichever is appropriate
    """

    return "daily" if timing == "day" else f"{timing}ly"


def get_job_state(job: Job) -> list[str]:
    row = job._str()
    return [
        row[0],
        row[1] + row[2],
        job.handle.__name__,
        row[3],
        str(job.datetime.tzinfo),
        row[5],
        f"{row[6]}/{row[7]}",
        ",".join(job.tags)
    ]


from typing import Annotated

from fastapi import APIRouter, Body, Path, Query, Response, status

from orchestra.api.dto import JobDTO, ScheduleDefinitionDTO, RunDTO
from orchestra.core import instance
from orchestra.models import Run

router = APIRouter()


@router.get("/jobs/", tags=["jobs"])
async def get_jobs(tags: Annotated[str | None, Query(description="Comma separated tags to filter scheduled jobs")] = None,
                   any_tag: Annotated[bool, Query(description="Set to true to match if only at least one tag has to match")] = False,
                   include_paused: Annotated[bool, Query(description="Set to true to include paused jobs when searching for a match")] = True) -> list[JobDTO]:
    tags: set[str] = set(tags.split(",")) if tags else set()
    return list(map(lambda job: JobDTO.map(job), instance.get_jobs(tags, any_tag, include_paused=include_paused)))


@router.post("/jobs/", tags=["jobs"], status_code=201)
async def schedule_new_job(module: Annotated[str, Body()], schedule_definition: ScheduleDefinitionDTO):
    module_definition = [
        {"module": module,
         "schedules": [
             {
                 "name": schedule_definition.name,
                 "task": schedule_definition.task,
                 "enabled": True,
                 "schedule": {
                     "timing": schedule_definition.timing,
                     "timezone": schedule_definition.timezone or "utc"
                 }
             }
         ]}]
    instance.add_schedules(module_definition, attempt_resume=schedule_definition.resume or False)


@router.delete("/jobs/", tags=["jobs"])
async def delete_scheduled_jobs_matching_tags(tags: Annotated[str | None, Query(description="Comma separated tags to filter scheduled jobs")] = None,
                                              any_tag: Annotated[bool, Query(description="Set to true to match if only at least one tag has to match")] = False,
                                              include_paused: Annotated[bool, Query(description="Set to true to include paused jobs when searching for a match")] = False) -> list[JobDTO]:
    tags: set[str] = set(tags.split(",")) if tags else set()
    return list(map(lambda job: JobDTO.map(job), instance.delete_jobs_with_tags(tags, any_tag, include_paused)))


@router.post("/jobs/trigger/", tags=["jobs"])
async def trigger_scheduled_jobs_matching_tags(tags: Annotated[str | None, Query(description="Comma separated tags to filter scheduled jobs")] = None,
                                               any_tag: Annotated[bool, Query(description="Set to true to match if only at least one tag has to match")] = True,
                                               include_paused: Annotated[bool, Query(description="Set to true to include paused jobs when searching for a match")] = True) -> list[JobDTO]:
    tags: set[str] = set(tags.split(",")) if tags else set()
    return list(map(lambda job: JobDTO.map(job), await instance.trigger_jobs_with_tags(tags, any_tag, include_paused)))


@router.post("/jobs/pause/", tags=["jobs"])
async def pause_scheduled_jobs_matching_tags(tags: Annotated[str | None, Query(description="Comma separated tags to filter scheduled jobs")] = None,
                                             any_tag: Annotated[bool, Query(description="Set to true to match if only at least one tag has to match")] = True) -> list[JobDTO]:
    tags: set[str] = set(tags.split(",")) if tags else set()
    return list(map(lambda job: JobDTO.map(job), instance.pause_jobs_with_tags(tags, any_tag)))


@router.post("/jobs/resume/", tags=["jobs"])
async def resume_scheduled_jobs_matching_tags(tags: Annotated[str | None, Query(description="Comma separated tags to filter scheduled jobs")] = None,
                                              any_tag: Annotated[bool, Query(description="Set to true to match if only at least one tag has to match")] = True) -> list[JobDTO]:
    tags: set[str] = set(tags.split(",")) if tags else set()
    return list(map(lambda job: JobDTO.map(job), instance.resume_jobs_with_tags(tags, any_tag)))


@router.get("/jobs/{job_name}/runs", tags=["jobs"], status_code=status.HTTP_200_OK)
async def get_runs_of_job(job_name: Annotated[str, Path(description="The name of the job to for which to fetch runs")], response: Response, page_size: int = 50, page: int = 1) -> list[RunDTO] | str:
    if instance.job_exists(job_name):
        return list(map(lambda run: RunDTO.map(run), instance.get_runs_of_a_job(job_name, page_size, page)))
    else:
        response.status_code = status.HTTP_404_NOT_FOUND
        return f"Job {job_name} does not exist"


@router.get("/jobs/{job_name}/runs/{run_id}", tags=["jobs"], status_code=status.HTTP_200_OK)
async def get_run_of_job_by_id(job_name: Annotated[str, Path(description="The name of the job to for which to fetch runs")],
                               run_id: Annotated[int, Path(description="The id of the run to fetch")], response: Response) -> RunDTO | str:
    if instance.job_exists(job_name):
        run: Run | None = instance.get_run_of_a_job_by_id(job_name, run_id)
        if run is not None:
            return RunDTO.map(run)
        else:
            response.status_code = status.HTTP_404_NOT_FOUND
            return f"Run with run id {run_id} for job {job_name} does not exist"
    else:
        response.status_code = status.HTTP_404_NOT_FOUND
        return f"Job {job_name} does not exist"


@router.post("/jobs/{job_name}/trigger/", tags=["jobs"])
async def trigger_job_by_name(job_name: Annotated[str, Path(description="The name of the job to trigger")], response: Response) -> None | str:
    if instance.job_exists(job_name):
        response.status_code = status.HTTP_204_NO_CONTENT
        await instance.trigger_job(job_name)
    else:
        response.status_code = status.HTTP_404_NOT_FOUND
        return f"Job {job_name} does not exist"


@router.put("/jobs/{job_name}/pause", tags=["jobs"], status_code=status.HTTP_200_OK)
async def pause_job(job_name: Annotated[str, Path(description="The name of the job to pause")], response: Response) -> JobDTO | str:
    if instance.job_exists(job_name):
        return JobDTO.map(instance.pause_job(job_name))
    else:
        response.status_code = status.HTTP_404_NOT_FOUND
        return f"Job {job_name} does not exist"


@router.put("/jobs/{job_name}/resume", tags=["jobs"], status_code=status.HTTP_200_OK)
async def resume_job(job_name: Annotated[str, Path(description="The name of the job to resume")], response: Response) -> JobDTO | str:
    if instance.job_exists(job_name):
        return JobDTO.map(instance.resume_job(job_name))
    else:
        response.status_code = status.HTTP_404_NOT_FOUND
        return f"Job {job_name} does not exist"


@router.delete("/jobs/{job_name}", tags=["jobs"], status_code=status.HTTP_200_OK)
async def delete_scheduled_job(job_name: Annotated[str, Path(description="The name of the job to delete")], response: Response) -> JobDTO | str:
    if instance.job_exists(job_name):
        return JobDTO.map(instance.delete_job(job_name))
    else:
        response.status_code = status.HTTP_404_NOT_FOUND
        return f"Job {job_name} does not exist"



from orchestra.api.dto.job import JobDTO
from orchestra.api.dto.run import RunDTO
from orchestra.api.dto.task import TaskDTO
from orchestra.api.dto.schedule_definition import ScheduleDefinitionDTO

__all__ = ["JobDTO", "RunDTO", "TaskDTO", "ScheduleDefinitionDTO"]



from pydantic import BaseModel


class ScheduleDefinitionDTO(BaseModel):
    name: str
    task: str
    timing: str
    timezone: str | None = None
    tags: list[str] | None = None
    resume: bool = False



import datetime
from pydantic import BaseModel

from orchestra.models import Status, Run


class RunDTO(BaseModel):
    id: int
    job_name: str
    status: Status
    module: str
    task: str
    task_id: str
    schedule: str
    timezone: str | None
    triggered_date: datetime.datetime

    @classmethod
    def map(cls, run: Run) -> "RunDTO":
        status = Status[run.task_object.status.lower()] if run.task_object is not None else Status.pending
        return RunDTO(id=run.id,
                      job_name=run.job,
                      status=status,
                      module=run.module,
                      task=run.task,
                      task_id=run.task_id,
                      schedule=run.schedule.lower(),
                      timezone=run.timezone,
                      triggered_date=run.triggered_date)



import logging
import os
import signal

from fastapi import FastAPI
from rich.logging import RichHandler

from orchestra.api.routers import jobs, tasks
from orchestra.core import instance

logging.basicConfig(
    level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler(rich_tracebacks=True, tracebacks_show_locals=True)]
)

logger = logging.getLogger("orchestra.api")

app = FastAPI(title="Orchestra")
app.include_router(jobs.router)
app.include_router(tasks.router)

if instance is None:
    logger.error("Orchestra is not started, can't start API server")
    os.kill(os.getpid(), signal.SIGTERM)
